# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balpy',
 'balpy.balancerv2cad.scripts',
 'balpy.balancerv2cad.src.balancerv2cad',
 'balpy.balancerv2cad.src.balancerv2cad.logger',
 'balpy.balancerv2cad.tests',
 'balpy.balancerv2cad.tests.unit-test',
 'balpy.enums',
 'balpy.graph']

package_data = \
{'': ['*'],
 'balpy': ['abi/*',
           'balancerv2cad/*',
           'balancerv2cad/notebooks/*',
           'deployments/20210418-authorizer/abi/*',
           'deployments/20210418-authorizer/output/*',
           'deployments/20210418-vault/abi/*',
           'deployments/20210418-vault/output/*',
           'deployments/20210418-weighted-pool/abi/*',
           'deployments/20210418-weighted-pool/output/*',
           'deployments/20210624-stable-pool/abi/*',
           'deployments/20210624-stable-pool/output/*',
           'deployments/20210721-liquidity-bootstrapping-pool/abi/*',
           'deployments/20210721-liquidity-bootstrapping-pool/output/*',
           'deployments/20210727-meta-stable-pool/abi/*',
           'deployments/20210727-meta-stable-pool/output/*',
           'deployments/20210811-ldo-merkle/abi/*',
           'deployments/20210811-ldo-merkle/output/*',
           'deployments/20210812-lido-relayer/abi/*',
           'deployments/20210812-lido-relayer/output/*',
           'deployments/20210812-wsteth-rate-provider/abi/*',
           'deployments/20210812-wsteth-rate-provider/output/*',
           'deployments/20210907-investment-pool/abi/*',
           'deployments/20210907-investment-pool/output/*',
           'deployments/20210913-bal-arbitrum-merkle/abi/*',
           'deployments/20210913-bal-arbitrum-merkle/output/*',
           'deployments/20210928-mcb-arbitrum-merkle/abi/*',
           'deployments/20210928-mcb-arbitrum-merkle/output/*',
           'deployments/20211012-merkle-orchard/abi/*',
           'deployments/20211012-merkle-orchard/output/*',
           'deployments/20211202-no-protocol-fee-lbp/abi/*',
           'deployments/20211202-no-protocol-fee-lbp/output/*',
           'deployments/20211203-batch-relayer/abi/*',
           'deployments/20211203-batch-relayer/output/*',
           'deployments/20211208-aave-linear-pool/abi/*',
           'deployments/20211208-aave-linear-pool/output/*',
           'deployments/20211208-stable-phantom-pool/abi/*',
           'deployments/20211208-stable-phantom-pool/output/*',
           'deployments/20220304-erc4626-linear-pool/abi/*',
           'deployments/20220304-erc4626-linear-pool/output/*',
           'deployments/20220318-batch-relayer-v2/abi/*',
           'deployments/20220318-batch-relayer-v2/output/*',
           'deployments/20220325-authorizer-adaptor/abi/*',
           'deployments/20220325-authorizer-adaptor/output/*',
           'deployments/20220325-bal-token-holder-factory/abi/*',
           'deployments/20220325-bal-token-holder-factory/output/*',
           'deployments/20220325-balancer-token-admin/abi/*',
           'deployments/20220325-balancer-token-admin/output/*',
           'deployments/20220325-gauge-adder/abi/*',
           'deployments/20220325-gauge-adder/output/*',
           'deployments/20220325-gauge-controller/abi/*',
           'deployments/20220325-gauge-controller/output/*',
           'deployments/20220325-mainnet-gauge-factory/abi/*',
           'deployments/20220325-mainnet-gauge-factory/output/*',
           'deployments/20220325-single-recipient-gauge-factory/abi/*',
           'deployments/20220325-single-recipient-gauge-factory/output/*',
           'deployments/20220325-test-balancer-token/abi/*',
           'deployments/20220325-test-balancer-token/output/*',
           'deployments/20220325-ve-delegation/abi/*',
           'deployments/20220325-ve-delegation/output/*',
           'deployments/20220325-veBAL-deployment-coordinator/abi/*',
           'deployments/20220325-veBAL-deployment-coordinator/output/*',
           'deployments/20220413-arbitrum-root-gauge-factory/abi/*',
           'deployments/20220413-arbitrum-root-gauge-factory/output/*',
           'deployments/20220413-child-chain-gauge-factory/abi/*',
           'deployments/20220413-child-chain-gauge-factory/output/*',
           'deployments/20220413-polygon-root-gauge-factory/abi/*',
           'deployments/20220413-polygon-root-gauge-factory/output/*',
           'deployments/20220415-veBAL-L2-gauge-setup-coordinator/abi/*',
           'deployments/20220415-veBAL-L2-gauge-setup-coordinator/output/*',
           'deployments/20220418-veBAL-gauge-fix-coordinator/abi/*',
           'deployments/20220418-veBAL-gauge-fix-coordinator/output/*',
           'deployments/20220420-fee-distributor/abi/*',
           'deployments/20220420-fee-distributor/output/*',
           'deployments/20220420-smart-wallet-checker/abi/*',
           'deployments/20220420-smart-wallet-checker/output/*',
           'deployments/20220421-smart-wallet-checker-coordinator/abi/*',
           'deployments/20220421-smart-wallet-checker-coordinator/output/*',
           'deployments/20220422-distribution-scheduler/abi/*',
           'deployments/20220422-distribution-scheduler/output/*',
           'deployments/20220609-stable-pool-v2/abi/*',
           'deployments/20220609-stable-pool-v2/output/*',
           'deployments/20220725-protocol-fee-percentages-provider/abi/*',
           'deployments/20220725-protocol-fee-percentages-provider/output/*',
           'deployments/20220817-aave-rebalanced-linear-pool/abi/*',
           'deployments/20220817-aave-rebalanced-linear-pool/output/*',
           'deployments/20220906-composable-stable-pool/abi/*',
           'deployments/20220906-composable-stable-pool/output/*',
           'deployments/20220908-weighted-pool-v2/abi/*',
           'deployments/20220908-weighted-pool-v2/output/*',
           'deployments/20221021-managed-pool/abi/*',
           'deployments/20221021-managed-pool/output/*',
           'deployments/20230206-composable-stable-pool-v3/abi/*',
           'deployments/20230206-composable-stable-pool-v3/output/*',
           'deployments/20230206-erc4626-linear-pool-v3/abi/*',
           'deployments/20230206-erc4626-linear-pool-v3/output/*',
           'deployments/20230206-weighted-pool-v3/abi/*',
           'deployments/20230206-weighted-pool-v3/output/*']}

install_requires = \
['Cython==0.29.24',
 'cytoolz==0.11.2',
 'eth-abi==2.1.1',
 'gql==2.0.0',
 'jstyleson==0.0.2',
 'multicaller>=0.0.0a10',
 'requests==2.25.1',
 'web3==5.19.0']

setup_kwargs = {
    'name': 'balpy',
    'version': '0.0.0a81',
    'description': 'Balancer V2 Python API',
    'long_description': None,
    'author': 'Greg',
    'author_email': 'greg@orb.cl',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
